"use strict";
/**
 * GitDocumentDB plugin for remote connection using NodeGit
 * Copyright (c) Hidekazu Kubota
 *
 * This source code is licensed under the Mozilla Public License Version 2.0
 * found in the LICENSE file in the root directory of this source tree.
 */
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.createCredential = void 0;
const nodegit_1 = __importDefault(require("@sosuisen/nodegit"));
const error_1 = require("./error");
/**
 * Insert credential options for GitHub
 *
 * @internal
 */
function createCredentialForGitHub(options) {
    if (!options.remoteUrl.match(/^https?:\/\//)) {
        throw new error_1.Err.HttpProtocolRequiredError(options.remoteUrl);
    }
    const connection = options.connection;
    if (options.syncDirection !== 'pull' && !connection.personalAccessToken) {
        throw new error_1.Err.RemoteUndefinedPersonalAccessTokenError();
    }
    const urlArray = options.remoteUrl.replace(/^https?:\/\//, '').split('/');
    // github.com/account_name/repository_name
    if (urlArray.length !== 3) {
        throw new error_1.Err.InvalidRepositoryURLError(options.remoteUrl);
    }
    const owner = urlArray[urlArray.length - 2];
    const credentials = () => {
        return nodegit_1.default.Cred.userpassPlaintextNew(owner, connection.personalAccessToken);
    };
    return credentials;
}
/**
 * Create credential options for SSH
 *
 * @internal
 */
function createCredentialForSSH(options) {
    var _a;
    const connection = options.connection;
    if (connection.privateKeyPath === undefined || connection.privateKeyPath === '') {
        throw new error_1.Err.InvalidSSHKeyPathError();
    }
    if (connection.publicKeyPath === undefined || connection.publicKeyPath === '') {
        throw new error_1.Err.InvalidSSHKeyPathError();
    }
    (_a = connection.passPhrase) !== null && _a !== void 0 ? _a : (connection.passPhrase = '');
    const credentials = (url, userName) => {
        return nodegit_1.default.Cred.sshKeyNew(userName, connection.publicKeyPath, connection.privateKeyPath, connection.passPhrase);
    };
    return credentials;
}
/**
 * Create credential options
 *
 * @internal
 */
function createCredential(options) {
    var _a;
    (_a = options.connection) !== null && _a !== void 0 ? _a : (options.connection = { type: 'none' });
    let cred;
    if (options.connection.type === 'github') {
        cred = createCredentialForGitHub(options);
    }
    else if (options.connection.type === 'ssh') {
        cred = createCredentialForSSH(options);
    }
    else if (options.connection.type === 'none') {
        // nop
    }
    else {
        // @ts-ignore
        throw new error_1.Err.InvalidAuthenticationTypeError(options.connection.type);
    }
    const callbacks = {
        credentials: cred,
    };
    if (process.platform === 'darwin') {
        // @ts-ignore
        callbacks.certificateCheck = () => 0;
    }
    return callbacks;
}
exports.createCredential = createCredential;
//# sourceMappingURL=authentication.js.map