"use strict";
/**
 * GitDocumentDB plugin for remote connection using NodeGit
 * Copyright (c) Hidekazu Kubota
 *
 * This source code is licensed under the Mozilla Public License Version 2.0
 * found in the LICENSE file in the root directory of this source tree.
 */
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.type = exports.name = exports.push = exports.fetch = exports.clone = exports.checkPush = exports.checkFetch = exports.calcDistance = exports.sleep = void 0;
const nodegit_1 = __importDefault(require("@sosuisen/nodegit"));
const isomorphic_git_1 = __importDefault(require("isomorphic-git"));
const fs_1 = __importDefault(require("fs"));
const tslog_1 = require("tslog");
const error_1 = require("./error");
const const_1 = require("./const");
const authentication_1 = require("./authentication");
const net_1 = require("./net");
function sleep(msec) {
    return new Promise(resolve => setTimeout(resolve, msec));
}
exports.sleep = sleep;
const type = 'remote';
exports.type = type;
const name = 'nodegit';
exports.name = name;
/**
 * Clone
 *
 * @throws {@link Err.CannotConnectError}
 * @throws {@link Err.CannotCloneRepositoryError}
 */
async function clone(workingDir, remoteOptions, logger) {
    logger !== null && logger !== void 0 ? logger : (logger = new tslog_1.Logger({
        name: 'plugin-nodegit',
        minLevel: 'trace',
        displayDateTime: false,
        displayFunctionName: false,
        displayFilePath: 'hidden',
    }));
    if (remoteOptions !== undefined &&
        remoteOptions.remoteUrl !== undefined &&
        remoteOptions.remoteUrl !== '') {
        /**
         * Retry if network errors.
         */
        let result = {
            ok: false,
        };
        let retry = 0;
        for (; retry < const_1.NETWORK_RETRY; retry++) {
            // eslint-disable-next-line no-await-in-loop
            result = await net_1.checkHTTP(remoteOptions.remoteUrl, const_1.NETWORK_TIMEOUT).catch(err => err);
            if (result.ok) {
                break;
            }
            else {
                logger.debug(`NetworkError in cloning: ${remoteOptions.remoteUrl}, ` + result);
            }
            // eslint-disable-next-line no-await-in-loop
            await sleep(const_1.NETWORK_RETRY_INTERVAL);
        }
        if (!result.ok) {
            // Set retry number for code test
            throw new error_1.Err.CannotConnectError(retry, remoteOptions.remoteUrl, result.toString());
        }
        await nodegit_1.default.Clone.clone(remoteOptions.remoteUrl, workingDir, {
            fetchOpts: {
                callbacks: authentication_1.createCredential(remoteOptions),
            },
        }).catch(err => {
            // Errors except CannotConnectError are handled in combine functions.
            logger.debug(`Error in cloning: ${remoteOptions.remoteUrl}, ` + err);
            throw new error_1.Err.CannotCloneRepositoryError(remoteOptions.remoteUrl);
        });
    }
}
exports.clone = clone;
/**
 * Get or create Git remote named 'origin'
 *
 * (git remote add)
 *
 * @internal
 */
// eslint-disable-next-line complexity
async function _getOrCreateGitRemote(repos, remoteURL) {
    let result;
    // Check if remote repository already exists
    let remote = await nodegit_1.default.Remote.lookup(repos, 'origin').catch(() => { });
    if (remote === undefined) {
        // Add remote repository
        remote = await nodegit_1.default.Remote.create(repos, 'origin', remoteURL);
        result = 'add';
    }
    else if (remote.url() !== remoteURL) {
        nodegit_1.default.Remote.setUrl(repos, 'origin', remoteURL);
        result = 'change';
    }
    else {
        result = 'exist';
    }
    return [result, remote];
}
/**
 * Check connection by FETCH
 *
 * @throws {@link Err.InvalidURLError}
 * @throws {@link Err.RemoteRepositoryNotFoundError}
 * @throws Error (Other errors from NodeGit.Remote#connect())
 *
 * @internal
 */
// eslint-disable-next-line complexity
async function checkFetch(workingDir, options, logger) {
    var _a;
    logger !== null && logger !== void 0 ? logger : (logger = new tslog_1.Logger({
        name: 'plugin-nodegit',
        minLevel: 'trace',
        displayDateTime: false,
        displayFunctionName: false,
        displayFilePath: 'hidden',
    }));
    const credential = authentication_1.createCredential(options);
    const repos = await nodegit_1.default.Repository.open(workingDir);
    // Get NodeGit.Remote
    const [gitResult, remote] = await _getOrCreateGitRemote(repos, options.remoteUrl);
    logger.debug('Git remote: ' + gitResult);
    const remoteURL = remote.url();
    const error = String(await remote
        .connect(0 /* FETCH */, credential)
        .catch(err => err));
    await remote.disconnect();
    // if (error !== 'undefined') console.warn('connect fetch error: ' + error);
    switch (true) {
        case error === 'undefined':
            break;
        case error.startsWith('Error: unsupported URL protocol'):
        case error.startsWith('Error: failed to resolve address'):
        case error.startsWith('Error: failed to send request'):
            throw new error_1.Err.InvalidURLError(remoteURL + ':' + error);
        case error.startsWith('Error: unexpected HTTP status code: 4'): // 401, 404 on Ubuntu
        case error.startsWith('Error: request failed with status code: 4'): // 401, 404 on Windows
        case error.startsWith('Error: Method connect has thrown an error'):
        case error.startsWith('Error: ERROR: Repository not found'):
            // Remote repository does not exist, or you do not have permission to the private repository
            if (((_a = options.connection) === null || _a === void 0 ? void 0 : _a.type) === 'github') {
                return 'not_exist';
            }
            throw new error_1.Err.RemoteRepositoryNotFoundError(remoteURL + ':' + error);
        case error.startsWith('Error: remote credential provider returned an invalid cred type'): // on Ubuntu
        case error.startsWith('Failed to retrieve list of SSH authentication methods'):
        case error.startsWith('Error: too many redirects or authentication replays'):
            throw new error_1.Err.FetchPermissionDeniedError(error);
        default:
            throw new Error(error);
    }
    ;
    return 'exist';
}
exports.checkFetch = checkFetch;
/**
 * Check connection by PUSH
 *
 * @throws {@link Err.InvalidURLError}
 * @throws {@link Err.RemoteRepositoryNotFoundError}
 * @throws {@link Err.PushPermissionDeniedError}
 * @throws Error (Other errors from NodeGit.Remote#connect())
 *
 * @internal
 */
// eslint-disable-next-line complexity
async function checkPush(workingDir, options, logger) {
    logger !== null && logger !== void 0 ? logger : (logger = new tslog_1.Logger({
        name: 'plugin-nodegit',
        minLevel: 'trace',
        displayDateTime: false,
        displayFunctionName: false,
        displayFilePath: 'hidden',
    }));
    const credential = authentication_1.createCredential(options);
    const repos = await nodegit_1.default.Repository.open(workingDir);
    // Get NodeGit.Remote
    const [gitResult, remote] = await _getOrCreateGitRemote(repos, options.remoteUrl);
    logger.debug('Git remote: ' + gitResult);
    const error = String(await remote
        .connect(1 /* PUSH */, credential)
        .catch(err => err));
    await remote.disconnect();
    // if (error !== 'undefined') console.warn('connect push error: ' + error);
    switch (true) {
        case error === 'undefined':
            break;
        case error.startsWith('Error: unsupported URL protocol'):
        case error.startsWith('Error: failed to resolve address'):
        case error.startsWith('Error: failed to send request'):
            throw new error_1.Err.InvalidURLError(options.remoteUrl);
        case error.startsWith('Error: unexpected HTTP status code: 4'): // 401, 404 on Ubuntu
        case error.startsWith('Error: request failed with status code: 4'): // 401, 404 on Windows
        case error.startsWith('Error: Method connect has thrown an error'):
        case error.startsWith('Error: ERROR: Repository not found'): {
            // Remote repository does not exist, or you do not have permission to the private repository
            throw new error_1.Err.RemoteRepositoryNotFoundError(options.remoteUrl);
        }
        // Invalid personal access token
        // Personal access token is read only
        case error.startsWith('Error: remote credential provider returned an invalid cred type'): // on Ubuntu
        case error.startsWith('Error: too many redirects or authentication replays'):
        case error.startsWith('Error: ERROR: Permission to'): {
            throw new error_1.Err.PushPermissionDeniedError(error);
        }
        default:
            throw new Error(error);
    }
    return 'ok';
}
exports.checkPush = checkPush;
/**
 * Calc distance
 */
function calcDistance(baseCommitOid, localCommitOid, remoteCommitOid) {
    if (baseCommitOid === undefined) {
        return {
            ahead: undefined,
            behind: undefined,
        };
    }
    return {
        ahead: localCommitOid !== baseCommitOid ? 1 : 0,
        behind: remoteCommitOid !== baseCommitOid ? 1 : 0,
    };
}
exports.calcDistance = calcDistance;
/**
 * git push
 *
 * @throws {@link Err.UnfetchedCommitExistsError} (from this and validatePushResult())
 * @throws {@link Err.GitFetchError} (from validatePushResult())
 * @throws {@link Err.GitPushError} (from NodeGit.Remote.push())
 */
async function push(workingDir, remoteOptions) {
    const repos = await nodegit_1.default.Repository.open(workingDir);
    const remote = await repos.getRemote('origin');
    const credential = authentication_1.createCredential(remoteOptions);
    await remote
        .push(['refs/heads/main:refs/heads/main'], {
        callbacks: credential,
    })
        .catch((err) => {
        if (err.message.startsWith('cannot push because a reference that you are trying to update on the remote contains commits that are not present locally')) {
            throw new error_1.Err.UnfetchedCommitExistsError();
        }
        throw new error_1.Err.GitPushError(err.message);
    });
    await validatePushResult(repos, workingDir, credential);
    // It leaks memory if not cleanup
    repos.cleanup();
}
exports.push = push;
/**
 * NodeGit.Remote.push does not return valid error in race condition,
 * so check is needed.
 *
 * @throws {@link Err.GitFetchError}
 * @throws {@link Err.UnfetchedCommitExistsError}
 */
async function validatePushResult(repos, workingDir, credential) {
    await repos
        .fetch('origin', {
        callbacks: credential,
    })
        .catch(err => {
        throw new error_1.Err.GitFetchError(err.message);
    });
    // Use isomorphic-git to avoid memory leak
    const localCommitOid = await isomorphic_git_1.default.resolveRef({
        fs: fs_1.default,
        dir: workingDir,
        ref: 'HEAD',
    });
    const remoteCommitOid = await isomorphic_git_1.default.resolveRef({
        fs: fs_1.default,
        dir: workingDir,
        ref: 'refs/remotes/origin/main',
    });
    const [baseCommitOid] = await isomorphic_git_1.default.findMergeBase({
        fs: fs_1.default,
        dir: workingDir,
        oids: [localCommitOid, remoteCommitOid],
    });
    const distance = await calcDistance(baseCommitOid, localCommitOid, remoteCommitOid);
    if (distance.behind === undefined) {
        // This will not be occurred.
        throw new error_1.Err.NoMergeBaseFoundError();
    }
    else if (distance.behind > 0) {
        throw new error_1.Err.UnfetchedCommitExistsError();
    }
}
/**
 * git fetch
 *
 * @throws {@link Err.GitFetchError}
 */
async function fetch(workingDir, remoteOptions, logger) {
    logger !== null && logger !== void 0 ? logger : (logger = new tslog_1.Logger({
        name: 'plugin-nodegit',
        minLevel: 'trace',
        displayDateTime: false,
        displayFunctionName: false,
        displayFilePath: 'hidden',
    }));
    logger.debug(`sync_worker: fetch: ${remoteOptions.remoteUrl}`);
    const repos = await nodegit_1.default.Repository.open(workingDir);
    const credential = authentication_1.createCredential(remoteOptions);
    await repos.fetch('origin', {
        callbacks: credential,
    })
        .catch(err => {
        throw new error_1.Err.GitFetchError(err.message);
    });
    // It leaks memory if not cleanup    
    repos.cleanup();
}
exports.fetch = fetch;
//# sourceMappingURL=remote-nodegit.js.map